@@ -29,7 +29,6 @@
 #include <errno.h>
 
 #include <ell/ell.h>
-#include <ell/dbus-private.h>
 
 #include "linux/nl80211.h"
 
@@ -48,6 +47,12 @@ const char *dbus_iftype_to_string(uint32_t iftype)
 		return "station";
 	case NL80211_IFTYPE_AP:
 		return "ap";
+	case NL80211_IFTYPE_P2P_CLIENT:
+		return "p2p-client";
+	case NL80211_IFTYPE_P2P_GO:
+		return "p2p-go";
+	case NL80211_IFTYPE_P2P_DEVICE:
+		return "p2p-device";
 	default:
 		break;
 	}
