@@ -27,12 +27,16 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <errno.h>
+#include <alloca.h>
 #include <ell/ell.h>
 
-#include "crypto.h"
-#include "simutil.h"
-#include "simauth.h"
-#include "util.h"
+#include "src/missing.h"
+#include "src/eap.h"
+#include "src/eap-private.h"
+#include "src/crypto.h"
+#include "src/simutil.h"
+#include "src/simauth.h"
+#include "src/util.h"
 
 /*
  * EAP-SIM authentication protocol.
@@ -120,6 +124,15 @@ struct eap_sim_handle {
 	unsigned int auth_watch;
 };
 
+static void eap_sim_clear_secrets(struct eap_sim_handle *sim)
+{
+	explicit_bzero(sim->mk, sizeof(sim->mk));
+	explicit_bzero(sim->k_encr, sizeof(sim->k_encr));
+	explicit_bzero(sim->k_aut, sizeof(sim->k_aut));
+	explicit_bzero(sim->msk, sizeof(sim->msk));
+	explicit_bzero(sim->emsk, sizeof(sim->emsk));
+}
+
 static void eap_sim_free(struct eap_state *eap)
 {
 	struct eap_sim_handle *sim = eap_get_data(eap);
@@ -127,6 +140,8 @@ static void eap_sim_free(struct eap_state *eap)
 	if (sim->auth)
 		sim_auth_unregistered_watch_remove(sim->auth, sim->auth_watch);
 
+	eap_sim_clear_secrets(sim);
+
 	l_free(sim->identity);
 	l_free(sim->vlist);
 	l_free(sim);
@@ -290,11 +305,10 @@ static void gsm_callback(const uint8_t *sres, const uint8_t *kc,
 	uint8_t *pos = response;
 	uint8_t prng_buf[160];
 	uint8_t *mac_pos;
+	bool r;
 
-	if (!sres || !kc) {
-		l_free(sim->chal_pkt);
+	if (!sres || !kc)
 		goto chal_error;
-	}
 
 	if (sim->protected)
 		resp_len += 4;
@@ -307,8 +321,11 @@ static void gsm_callback(const uint8_t *sres, const uint8_t *kc,
 
 	eap_sim_fips_prf(sim->mk, 20, prng_buf, 160);
 
-	if (!eap_sim_get_encryption_keys(prng_buf, sim->k_encr, sim->k_aut,
-			sim->msk, sim->emsk)) {
+	r = eap_sim_get_encryption_keys(prng_buf, sim->k_encr, sim->k_aut,
+					sim->msk, sim->emsk);
+	explicit_bzero(prng_buf, sizeof(prng_buf));
+
+	if (!r) {
 		l_error("could not derive encryption keys");
 		goto chal_fatal;
 	}
@@ -317,7 +334,6 @@ static void gsm_callback(const uint8_t *sres, const uint8_t *kc,
 			sim->pkt_len, sim->k_aut, sim->nonce,
 			EAP_SIM_NONCE_LEN)) {
 		l_error("server MAC was invalid");
-		l_free(sim->chal_pkt);
 		goto chal_error;
 	}
 
@@ -354,6 +370,7 @@ static void gsm_callback(const uint8_t *sres, const uint8_t *kc,
 	}
 
 	l_free(sim->chal_pkt);
+	sim->chal_pkt = NULL;
 
 	eap_send_response(eap, EAP_TYPE_SIM, response, resp_len);
 
@@ -362,7 +379,9 @@ static void gsm_callback(const uint8_t *sres, const uint8_t *kc,
 		 * Result indication not required, we must accept success.
 		 */
 		eap_method_success(eap);
-		eap_set_key_material(eap, sim->msk, 32, NULL, 0, NULL, 0);
+		/* TODO: Derive Session-ID */
+		eap_set_key_material(eap, sim->msk, 32, NULL, 0, NULL, 0,
+					NULL, 0);
 
 		sim->state = EAP_SIM_STATE_SUCCESS;
 	}
@@ -378,6 +397,8 @@ chal_fatal:
 	return;
 
 chal_error:
+	l_free(sim->chal_pkt);
+	sim->chal_pkt = NULL;
 	eap_sim_client_error(eap, EAP_TYPE_SIM, EAP_SIM_ERROR_PROCESS);
 }
 
@@ -447,6 +468,7 @@ static void handle_challenge(struct eap_state *eap, const uint8_t *pkt,
 
 	if (sim_auth_run_gsm(sim->auth, rands, 3, gsm_callback, eap) < 0) {
 		l_free(sim->chal_pkt);
+		sim->chal_pkt = NULL;
 		goto chal_error;
 	}
 
@@ -512,7 +534,9 @@ static void handle_notification(struct eap_state *eap, const uint8_t *pkt,
 		 * Server sent successful result indication
 		 */
 		eap_method_success(eap);
-		eap_set_key_material(eap, sim->msk, 32, NULL, 0, NULL, 0);
+		/* TODO: Derive Session-ID */
+		eap_set_key_material(eap, sim->msk, 32, NULL, 0, NULL, 0,
+					NULL, 0);
 
 		/*
 		 * Build response packet
@@ -630,6 +654,23 @@ static int eap_sim_check_settings(struct l_settings *settings,
 	return 0;
 }
 
+static bool eap_sim_reset_state(struct eap_state *eap)
+{
+	struct eap_sim_handle *sim = eap_get_data(eap);
+
+	sim->state = EAP_SIM_STATE_UNCONNECTED;
+
+	l_free(sim->vlist);
+	sim->vlist = NULL;
+	l_free(sim->chal_pkt);
+	sim->chal_pkt = NULL;
+
+	memset(sim->nonce, 0, sizeof(sim->nonce));
+	eap_sim_clear_secrets(sim);
+
+	return true;
+}
+
 static bool eap_sim_load_settings(struct eap_state *eap,
 					struct l_settings *settings,
 					const char *prefix)
@@ -667,7 +708,8 @@ static struct eap_method eap_sim = {
 	.handle_request = eap_sim_handle_request,
 	.check_settings = eap_sim_check_settings,
 	.load_settings = eap_sim_load_settings,
-	.get_identity = eap_sim_get_identity
+	.get_identity = eap_sim_get_identity,
+	.reset_state = eap_sim_reset_state
 };
 
 static int eap_sim_init(void)
